"""
The content of this library is based on the following publication:

[C. F. F. Karney (2013), "Algorithms for geodesics", J. Geodesy 87, 43-55](https://doi.org/10.1007/s00190-012-0578-z)

This paper should be cited in publications using this library.

The MIT License (MIT).

Copyright (c) 2021-{0}, Per A. Brodtkorb

Permission is hereby granted, free of charge, to any person
obtaining a copy of this software and associated documentation
files (the "Software"), to deal in the Software without
restriction, including without limitation the rights to use, copy,
modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""


from datetime import datetime
CURRENT_YEAR = datetime.now().year

if __doc__:
    __doc__ = __doc__.format(CURRENT_YEAR)  # @ReservedAssignment



"""
Geodesic module
---------------

"""
import warnings
import numpy as np
from numpy import arctan2, sin, cos, tan, arctan, sqrt
# from scipy.special import ellipeinc, ellipkinc  # pylint: disable=no-name-in-module
from karney._common import _make_summary, test_docstrings
from karney.util import (nthroot,
                         eccentricity2,
                         third_flattening,
                         polar_radius,
                         deg,
                         rad,
                         # EPS,
                         TINY)

__all__ = ["distance", "reckon", "sphere_distance_rad"]
# TINY = 1e-150

# A1 coefficients defined in eq. 17 in Karney:
A1_COEFFICIENTS = (1. / 256, 1. / 64., 1. / 4., 1.)
# C1 coefficients defined in eq. 18 in Karney:
C1_COEFFICIENTS = (
    (-1. / 32., 3. / 16., -1. / 2, ),  # C11
    (-9. / 2048., 1. / 32., -1. / 16.),  # C12
    (3. / 256, -1. / 48.),  # C13
    (3. / 512., -5. / 512.),  # C14
    (-7. / 1280.,),  # C15
    (-7. / 2048.,),  # C16
)
# CM1 coefficients defined in eq. 21 in Karney:
CM1_COEFFICIENTS = (
    (205. / 1536., -9. / 32., 1. / 2, ),  # CM11
    (1335. / 4096, -37. / 96., 5. / 16.),  # CM12
    (-75. / 128, 29. / 96.),  # CM13
    (-2391. / 2560., 539. / 1536.),  # CM14
    (3467. / 7680.,),  # CM15
    (38081. / 61440.,)  # CM16
)

# A2 coefficients defined in eq. 42 in Karney:
A2_COEFFICIENTS = (25. / 256., 9. / 64., 1./4., 1)
# C2 coefficients defined in eq. 43 in Karney:
C2_COEFFICIENTS = (
    (1. / 32., 1. / 16., 1./2.),  # C21
    (35. / 2048, 1./32., 3. / 16.),  # C22
    (5. / 256, 5. / 48.),  # C23
    (7. / 512., 35. / 512.),  # C24
    (63. / 1280.,),  # C25
    (77. / 2048.,),  # C26
)

# A3 coefficients defined in eq. 24 in Karney:
A3_COEFFICIENTS = (
    (-3. / 128.,),
    (-2. / 64., -3. / 64.),
    (-1. / 16., -3. / 16., -1. / 16.),
    (3. / 8., -1. / 8., -1. / 4.),
    (1. / 2., -1. / 2., ),
    (1., ))
# C3 coefficients defined in eq. 25 in Karney:
C3_COEFFICIENTS = (
    ((3, 128.), (2, 5, 128.), (-1, 3, 3, 64.), (-1, 0, 1, 8.), (-1, 1, 4.)),  # C_31
    ((5, 256.), (1, 3, 128.), (-3, -2, 3, 64.), (1, -3, 2, 32.)),  # C_32
    ((7, 512.), (-10, 9, 384.), (5, -9, 5, 192.)),  # C_33
    ((7, 512.), (-14, 7, 512.)),  # C_34
    ((21, 2560.),),  # C_35
)


def __astroid(x, y):
    """
    ASTROID  Solve the astroid equation

    K = ASTROID(X, Y) solves the quartic polynomial Eq. (55):

      K^4 + 2 * K^3 - (X^2 + Y^2 - 1) * K^2 - 2*Y^2 * K - Y^2 = 0

    for the positive root K.  X and Y are arrays of the same shape
    and the returned value K has the same shape.
    """
    x, y = np.atleast_1d(x, y)
    k = np.zeros(x.shape)
    p = x**2
    q = y**2
    r = (p + q - 1) / 6
    fl1 = ~((q == 0) & (r <= 0))
    p = p[fl1]
    q = q[fl1]
    r = r[fl1]
    pq4 = p * q / 4
    r_2 = r**2
    r_3 = r * r_2
    disc = pq4 * (pq4 + 2 * r_3)
    u = r
    fl2 = disc >= 0
    t_3 = pq4[fl2] + r_3[fl2]
    t_3 = t_3 + (1 - 2 * (t_3 < 0)) * sqrt(disc[fl2])
    t = np.sign(t_3)*nthroot(np.abs(t_3), 3)
    u[fl2] = u[fl2] + t + r_2[fl2] / np.where(t != 0, t, np.inf)
    ang = arctan2(sqrt(-disc[~fl2]), -(pq4[~fl2] + r_3[~fl2]))
    u[~fl2] = u[~fl2] + 2 * r[~fl2] * cos(ang / 3)
    v = sqrt(u**2 + q)
    u_v = u + v
    fl2 = u < 0
    u_v[fl2] = q[fl2] / (v[fl2] - u[fl2])
    w = (u_v - q) / (2 * v)
    k[fl1] = u_v / (sqrt(u_v + w**2) + w)
    return k


def _astroid(x, y, f):
    m_u = __astroid(x, y)
    # Oblate solution
    alpha1o = np.where(y == 0,
                       arctan2(-x, sqrt(np.maximum(1 - x**2, 0))),
                       arctan2(-x * m_u / (1 + m_u), y))  # Eq. 56 and 57
    # Prolate solution
    alpha1p = np.where(y == 0,
                       arctan2(sqrt(np.maximum(1 - x**2, 0)), -x),
                       arctan2(-y, x * m_u / (1 + m_u)))
    shape = alpha1o.shape
    alpha11 = np.where((f < 0)*np.ones(shape), alpha1p, alpha1o)

    return alpha11


def _eval_cij_coefs(coefficients, epsi, squared=True):
    epsi2 = epsi**2 if squared else epsi
    factor = 1.0
    c1x = []
    for coefs in coefficients:
        factor = factor * epsi
        c1x.append(factor*np.polyval(coefs, epsi2))
    return c1x


def _cosinesum(c, x, sine=True):
    """
    Returns the sum of the sine or cosine series using Clenshaw algorithm.

    Parameters
    ----------
    c : list
        sine or cosine series coefficients
    x : array-like
        argument to the sine or cosine series.
    sine: bool
        If True the sine series sum is returned otherwise the cosine series sum.

    Returns
    -------
    y : array-like
        If sine is True y = sum c[i-1] * sin( 2*i * x)
        otherwise y = sum c[i-1] * cos((2*i-1) * x) for i = 1, .... n
    """

    cosx, sinx = np.atleast_1d(cos(x), sin(x))
    n = len(c)
    factor = 2 * (cosx - sinx) * (cosx + sinx)
    y_1 = np.zeros(sinx.shape)
    is_odd = n % 2
    if is_odd:
        y_0 = c[-1]
        n = n - 1
    else:
        y_0 = y_1

    for k in range(n-1, -1, -2):
        y_1 = factor * y_0 - y_1 + c[k]
        y_0 = factor * y_1 - y_0 + c[k-1]

    if sine:
        return 2 * sinx * cosx * y_0
    return cosx * (y_0 - y_1)


def _a3_coefs(n):
    """
    Returns the A3 coefficients defined in Eq. 24 in Karney evaluated at n.

    Parameters
    ----------
    n: real scalar
        third flattening of the ellipsoid
    """
    return [np.polyval(c, n) for c in A3_COEFFICIENTS]


def _c3_coefs(n):
    """
    Returns the C3 coefficients defined in Eq. 25 in Karney evaluated at n.

    Parameters
    ----------
    n: real scalar
        third flattening of the ellipsoid
    """
    return [[np.polyval(c[:-1], n) / c[-1] for c in coefs]
            for coefs in C3_COEFFICIENTS]


def _get_i3_fun(epsi, n=None, a3_coefs=None, c3_coefs=None):
    """
    Returns the I3 integral function defined in equation 8 in Karney

    Parameters
    ----------
    epsi: array-like
        normalized equatorial azimuth
    n: real scalar
        third flattening of the ellipsoid

    Returns
    -------
    i3fun : callable
        Integral function I3(sigma)

    Notes
    -----
    The I3 integral is defined as
      I3(sigma) = int (2-f)/(1+(1-f)*sqrt(1+k^2*sin(x)^2) dx from 0 to sigma

    Here
    f is the flattening
    n = f/(2-f) is the third flattening
    e = sqrt(f*(2-f)) is the eccentricity
    em = e/sqrt(1-e^2) is the second eccentricity
    alpha0 is the equatorial azimuth
    k = em*cos(alpha0)
    epsi = (sqrt(1+k^2)-1)/(sqrt(1+k^2)+1)
    sigma is the spherical arc length of the auxiliary sphere


    References
    ----------
    C. F. F. Karney, "Algorithms for geodesics",
    J. Geodesy 87, 43-55 (2013);
    https://doi.org/10.1007/s00190-012-0578-z
    Addenda: https://geographiclib.sourceforge.io/geod-addenda.html

    """
    if n is not None:
        a3_coefs = _a3_coefs(n)
        c3_coefs = _c3_coefs(n)

    a_3 = np.polyval(a3_coefs, epsi)  # Eq. 24
    c3x = _eval_cij_coefs(c3_coefs, epsi, squared=False)  # Eq 25

    def i3fun(sigma):
        return a_3*(sigma + _cosinesum(c3x, sigma, sine=True))
    return i3fun


def _get_i1_fun(epsi, return_inverse=True):
    """
    Returns the I1 integral function defined in equation 7 in Karney

    Parameters
    ----------
    epsi: array-like
        normalized equatorial azimuth

    Returns
    -------
    i1fun : callable
        Integral function I1(sigma)

    Notes
    -----
    The I1 integral is defined as
      I1(sigma) = int sqrt(1+k^2*sin(x)^2) dx from 0 to sigma

    Here
    f is the flattening

    e = sqrt(f*(2-f)) is the eccentricity
    em = e/sqrt(1-e^2) is the second eccentricity
    alpha0 is the equatorial azimuth
    k = em*cos(alpha0)
    epsi = (sqrt(1+k^2)-1)/(sqrt(1+k^2)+1) = k^2/(sqrt(1+k^2)+1)^2
    sigma is the spherical arc length of the auxiliary sphere


    References
    ----------
    C. F. F. Karney, "Algorithms for geodesics",
    J. Geodesy 87, 43-55 (2013);
    https://doi.org/10.1007/s00190-012-0578-z
    Addenda: https://geographiclib.sourceforge.io/geod-addenda.html

    """

    a_1 = np.polyval(A1_COEFFICIENTS, epsi**2) / (1.0 - epsi)  # Eq 17
    c1x = _eval_cij_coefs(C1_COEFFICIENTS, epsi, squared=True)  # Eq 18

    def i1fun(sigma):
        """The I1 function"""
        return a_1 * (sigma + _cosinesum(c1x, sigma, sine=True))

    if not return_inverse:
        return i1fun

    cm1x = _eval_cij_coefs(CM1_COEFFICIENTS, epsi, squared=True)  # Eq. 21

    def invi1fun(sdb):
        """The inverse of I1 function"""
        tau = sdb / a_1
        return tau + _cosinesum(cm1x, tau, sine=True)

#     k2 = -4 * epsi / (1-epsi)**2
#
#     def i1fun(sigma):
#         """The I1 function"""
#         return ellipeinc(sigma, k2)
#
#     def invi1fun(sdb, maxiter=20, atol=1e-12):
#         """The inverse of I1 function"""
#         sigma = sdb / a_1
#
#         for _ in range(maxiter):
#             delta = (sdb - ellipeinc(sigma, k2)) / sqrt(1 - k2 * sin(sigma)**2)
#             sigma += delta
#             if np.all(np.abs(delta) < atol):
#                 break
#         return sigma

    return i1fun, invi1fun


def _get_jfun(epsi):
    epsi2 = epsi**2

    epsim1 = 1.0 - epsi
    a_1 = np.polyval(A1_COEFFICIENTS, epsi2) / epsim1  # Eq 17
    a_2 = np.polyval(A2_COEFFICIENTS, epsi2) * epsim1  # Eq 42
    a1m2 = a_1-a_2

    c1x = _eval_cij_coefs(C1_COEFFICIENTS, epsi, squared=True)  # Eq 18
    c2x = _eval_cij_coefs(C2_COEFFICIENTS, epsi, squared=True)  # Eq 43
    c1m2x = a_1*np.array(c1x) - a_2*np.array(c2x)

    def jfun(sigma):
        """The J function defined as I1(sigma)-I2(sigma)"""
        return a1m2 * sigma + _cosinesum(c1m2x, sigma, sine=True)

#     k2 = -4 * epsi / (1-epsi)**2
#
#     def jfun(sigma):
#         """The J function defined as I1(sigma) - I2(sigma)"""
#         return ellipeinc(sigma, k2) - ellipkinc(sigma, k2)

    return jfun


def truncate_small(x, small=0.06):
    """Truncate tiny values to zero"""
    y = np.where(np.abs(x) < small, small - (small - x), x)
    return np.where(x == 0, 0, y)


def normalize_angle(angle):
    """Normalize angle to range (-pi, pi]"""
    mask = np.isfinite(angle)
    nangle = np.copy(angle)
    nangle[mask] = np.mod(angle[mask]+np.pi, 2*np.pi)-np.pi
    return np.where(nangle <= -np.pi, np.pi, nangle)


def _normalize_equatorial_azimuth(cos_alpha0, e2m):
    """
    Normalize the equatorial azimuth, alpha0, given the second eccentricity squared, e2m.
    """
    k_2 = e2m * cos_alpha0 ** 2
    k_1 = sqrt(1 + k_2)
    epsi = k_2 / (k_1 + 1)**2  # Eq. 16
    return epsi


def _solve_triangle_nea_direct(lat1, alpha1, f):
    """Returns alpha0, sigma1, w1, cos(alpha0), sin(alpha0)"""
    blat1 = arctan((1 - f) * tan(truncate_small(lat1)))  # Eq. 6
    return _solve_triangle_nea(blat1, alpha1)


def _solve_triangle_nea(blat1, alpha1):
    cos_alpha1, sin_alpha1 = cos(alpha1), sin(alpha1)
    cos_blat1, sin_blat1 = cos(blat1)+TINY, sin(blat1)
    sin_alpha0 = sin_alpha1 * cos_blat1  # Eq. 5
    cos_alpha0 = np.abs(cos_alpha1 + 1j * sin_alpha1 * sin_blat1)
    # alpha0 is arctan2(sin_alpha0, cos_alpha0)  # Eq 10
    sigma1 = arctan2(sin_blat1, cos_alpha1 * cos_blat1)  # Eq 11
    w_1 = arctan2(sin_alpha0 * sin(sigma1), cos(sigma1))  # Eq 12
    return sigma1, w_1, cos_alpha0, sin_alpha0


def _solve_triangle_neb_direct(sigma2, cos_alpha0, sin_alpha0):
    """Returns alpha2, blat2, w_2"""
    cos_sigma2, sin_sigma2 = cos(sigma2), sin(sigma2)
    sin_blat2 = cos_alpha0 * sin_sigma2
    cos_blat2 = np.abs(cos_alpha0 * cos_sigma2 + 1j * sin_alpha0)
    w_2 = arctan2(sin_alpha0 * sin_sigma2, cos_sigma2)  # Eq. 12
    blat2 = arctan2(sin_blat2, cos_blat2)  # Eq. 13
    alpha2 = arctan2(sin_alpha0, cos_alpha0 * cos_sigma2)  # Eq. 14
    return alpha2, blat2, w_2


def _solve_triangle_neb(cos_blat1, cos_blat2, sin_blat2, sin_alpha0, alpha1):
    # sgn is -1 to make sure sigma2==pi for antipodal points on equator:
    sgn = np.where((sin_blat2 == 0) & (cos_blat1 == 1), -1, 1)
    cos_alpha2_cos_blat2 = sgn*np.sqrt(cos(alpha1)**2 * cos_blat1**2
                                       + (cos_blat2**2 - cos_blat1**2))

    sin_alpha2_cos_blat2 = sin(alpha1)*cos_blat1
    alpha2 = arctan2(sin_alpha2_cos_blat2, cos_alpha2_cos_blat2)  # stable at both 0 and pi/2 angle
    # alpha20 = np.arccos(cos_alpha2_cos_blat2 / cos_blat2)  # Eq 45. in Karney
    sigma2 = arctan2(sin_blat2, cos_alpha2_cos_blat2)  # Eq 11
    w_2 = np.sign(sigma2)*np.abs(arctan2(sin_alpha0 * sin(sigma2), cos(sigma2)))  # Eq 12

    return sigma2, w_2, alpha2


def sphere_distance_rad(lat1, lon1, lat2, lon2):
    """
    Returns surface distance between positions A and B as well as the azimuths.

    Parameters
    ----------
    lat1, lon1: real scalars or vectors of length m.
        latitude(s) and longitude(s) [rad] of position A.
    lat2, lon2: real scalars or vectors of length n.
        latitude(s) and longitude(s) [rad] of position B.

    Returns
    -------
    distance:  real scalars or vectors of length max(m,n).
        Surface distance [rad] from A to B on the sphere
    azimuth_a, azimuth_b: real scalars or vectors of length max(m,n).
        direction [rad] of line at position A and B relative to
        North, respectively.

    Notes
    -----
    Solves the inverse geodesic problem of finding the length and azimuths of the
    shortest geodesic between points specified by lat1, lon1, lat2, lon2 on a sphere.

    Examples
    --------
    **"Surface distance"**
    ----------------------

    Find the surface distance sAB (i.e. great circle distance) between two
    positions A and B. The heights of A and B are ignored, i.e. if they don't have
    zero height, we seek the distance between the points that are at the surface of
    the Earth, directly above/below A and B.  Use Earth radius 6371e3 m.
    Compare the results with exact calculations for the WGS-84 ellipsoid.

    In geodesy this is known as "The second geodetic problem" or
    "The inverse geodetic problem" for a sphere/ellipsoid.

    Solution for a sphere:
      >>> import numpy as np
      >>> from karney.geodesic import rad, sphere_distance_rad, distance

      >>> latlon_a = (88, 0)
      >>> latlon_b = (89, -170)
      >>> latlons = latlon_a + latlon_b
      >>> r_Earth = 6371e3  # m, mean Earth radius
      >>> s_AB = sphere_distance_rad(*rad(latlons))[0]*r_Earth
      >>> s_AB = distance(*latlons, a=r_Earth, f=0, degrees=True)[0] # or alternatively

      >>> "Ex5: Great circle = {:5.2f} km".format(s_AB / 1000)
      'Ex5: Great circle = 332.46 km'

    Exact solution for the WGS84 ellipsoid:
      >>> s_12, azi1, azi2 = distance(*latlons, degrees=True)
      >>> "Ellipsoidal distance = {:5.2f} km".format(s_12 / 1000)
      'Ellipsoidal distance = 333.95 km'


    See also
    --------
    distance

    """
    w = lon2 - lon1
    cos_b1, sin_b1 = cos(lat1), sin(lat1)
    cos_b2, sin_b2 = cos(lat2), sin(lat2)
    cos_w, sin_w = cos(w), sin(w)

    sin_a1 = cos_b2 * sin_w
    cos_a1 = cos_b1*sin_b2-sin_b1*cos_b2*cos_w
    sin_a2 = cos_b1 * sin_w
    cos_a2 = -cos_b2 * sin_b1 + sin_b2 * cos_b1 * cos_w
    cos_distance_rad = sin_b1*sin_b2+cos_b1*cos_b2*cos_w

    sin_distance_rad = np.hypot(sin_a1, cos_a1)

    azimuth_a = arctan2(sin_a1, cos_a1)  # Eq 49
    azimuth_b = arctan2(sin_a2, cos_a2)  # Eq 50
    distance_rad = arctan2(sin_distance_rad, cos_distance_rad)  # Eq 51
    return distance_rad, azimuth_a, azimuth_b


def reckon(lat_a, lon_a, distance, azimuth, a=6378137, f=1.0 / 298.257223563,
           long_unroll=False, degrees=False):
    """
    Returns position B computed from position A, distance and azimuth.

    Parameters
    ----------
    lat_a, lon_b: real scalars or vectors of length k.
        latitude(s) and longitude(s) of position A [rad or deg].
    distance: real scalar or vector of length m.
        ellipsoidal distance [m] between position A and B.
    azimuth: real scalar or vector of length n.
        azimuth [rad or deg] of line at position A.
    a: real scalar, default WGS-84 ellipsoid.
        Semi-major half axis of the Earth ellipsoid given in [m].
    f: real scalar, default WGS-84 ellipsoid.
        Flattening [no unit] of the Earth ellipsoid. If f==0 then spherical
        Earth with radius a is used in stead of WGS-84.
    long_unroll: bool
        Controls the treatment of longitude. If it is False then the lon2
        is reduced to the range [-180, 180) or [-pi, pi). If it is True, then (lon2 - lon1)
        determines how many times and in what sense the geodesic has encircled
        the ellipsoid.
    degrees: bool
        True if latitude, longitude and azimuth is given in degress.

    Returns
    -------
    lat_b, lon_b:  arrays of length max(k,m,n).
        latitude(s) and longitude(s) of position B [rad or deg].
    azimuth_b: real scalars or vectors of length max(k,m,n).
        azimuth [rad or deg] of line at position B.

    Notes
    -----
    This function solves the direct geodesic problem of finding the final point B and
    azimuth_b given the position of point A and the azimuth and distance from A to B.
    Keep :math:`|f| <= 1/50` for full double precision accuracy.

    Examples
    --------
    **"A and azimuth/distance to B"**
    ---------------------------------
    We have an initial position A, direction of travel given as an azimuth
    (bearing) relative to north (clockwise), and finally the
    distance to travel along a great circle given as sAB.
    Use Earth radius 6371e3 m to find the destination point B.

    In geodesy this is known as "The first geodetic problem" or
    "The direct geodetic problem" for a sphere.

      >>> import numpy as np
      >>> from karney.geodesic import reckon
      >>> lat, lon = 80, -90
      >>> msg = "Ex8, Destination: lat, lon = {:4.4f} deg, {:4.4f} deg"

    Greatcircle solution:
      >>> datum = dict(a=6371e3, f=0)
      >>> lat2, lon2, azi_b = reckon(lat, lon, distance=1000, azimuth=200, degrees=True, **datum)

      >>> msg.format(lat2, lon2)
      'Ex8, Destination: lat, lon = 79.9915 deg, -90.0177 deg'

      >>> np.allclose(azi_b, -160.0174292682187)
      True

    Exact solution:
      >>> lat_b, lon_b, azi_b = reckon(lat, lon, distance=1000, azimuth=200, degrees=True)
      >>> msg.format(lat_b, lon_b)
      'Ex8, Destination: lat, lon = 79.9916 deg, -90.0176 deg'

      >>> np.allclose(azi_b, -160.01742926820506)
      True


    See also
    --------
    distance
    """
    if degrees:
        params = rad(lat_a, lon_a) + (distance, rad(azimuth), a, f, long_unroll)
        return deg(*_reckon(*params))
    return _reckon(lat_a, lon_a, distance, azimuth, a, f, long_unroll)


def _reckon(lat_a, lon_a, distance, azimuth, a=6378137, f=1.0 / 298.257223563, long_unroll=False):
    """
    Returns position B computed from position A, distance and azimuth.

    Parameters
    ----------
    lat_a, lon_a: real scalars or vectors of length k.
        latitude(s) and longitude(s) of position A [rad].
    distance: real scalar or vector of length m.
        ellipsoidal distance [m] between position A and B.
    azimuth: real scalar or vector of length n.
        azimuth [rad] of line at position A.
    a: real scalar, default WGS-84 ellipsoid.
        Semi-major half axis of the Earth ellipsoid given in [m].
    f: real scalar, default WGS-84 ellipsoid.
        Flattening [no unit] of the Earth ellipsoid. If f==0 then spherical
        Earth with radius a is used in stead of WGS-84.
    long_unroll: bool
        Controls the treatment of longitude. If it is False then the lon2
        is reduced to the range [-pi, pi). If it is True, then (lon2 - lon1)
        determines how many times and in what sense the geodesic has encircled
        the ellipsoid.

    Returns
    -------
    lat_b, lon_b:  arrays of length max(k,m,n).
        latitude(s) and longitude(s) of position B [rad].
    azimuth_b: real scalars or vectors of length max(k,m,n).
        azimuth [rad] of line at position B.

    """

    lat1, lon1, distance, az1, a = np.broadcast_arrays(lat_a, lon_a, distance, azimuth, a)
    alpha1 = truncate_small(az1)
    sigma1, w_1, cos_alpha0, sin_alpha0 = _solve_triangle_nea_direct(lat1, alpha1, f)

    # Determine sigma2:
    n = third_flattening(f)
    e2m = eccentricity2(f)[1]
    epsi = _normalize_equatorial_azimuth(cos_alpha0, e2m)

    i1fun, i1inv = _get_i1_fun(epsi)
    b = polar_radius(a, f)
    s_1 = b * i1fun(sigma1)  # Eq. 7. I1(sigma1)
    s_2 = s_1 + distance
    sigma2 = i1inv(s_2/b)  # Eq. 20: Inverse of I1 where I1 is defined in Eq. 7.

    alpha2, blat2, w_2 = _solve_triangle_neb_direct(sigma2, cos_alpha0, sin_alpha0)

    # Determine lamda12
    fun_i3 = _get_i3_fun(epsi, n)
    # lamda1 is w_1 - f * sin_alpha0 * fun_i3(sigma1)  # Eq. 8
    # lamda2 is w_2 - f * sin_alpha0 * fun_i3(sigma2)  # Eq. 8

    lamda12 = w_2-w_1 + f * sin_alpha0 * (fun_i3(sigma1) - fun_i3(sigma2))

    if long_unroll:
        correction = (sigma2 - arctan2(sin(sigma2), cos(sigma2))
                      - sigma1 + arctan2(sin(sigma1), cos(sigma1)))
        sgn = np.where(sin_alpha0 >= 0, 1, -1)
        lon2 = lon1 + lamda12 + sgn * correction
    else:
        lon2 = normalize_angle(lon1 + lamda12)
    lat2 = arctan(tan(blat2)/(1-f))  # Eq. 6

    if np.ndim(lat1) == 0:
        return lat2[0], lon2[0], alpha2[0]
    return lat2, lon2, alpha2


def _solve_alpha1(alpha1, blat1, blat2, true_lamda12, a, f, tol=1e-15):
    b = polar_radius(a, f)
    eta = third_flattening(f)
    e_2, e2m = eccentricity2(f)

    sin_blat1, cos_blat1 = sin(blat1)-TINY, cos(blat1)
    sin_blat2, cos_blat2 = sin(blat2), cos(blat2)

    def _newton_step(alpha1):
        """ See table 5 in Karney"""
        sigma1, w_1, cos_alpha0, sin_alpha0 = _solve_triangle_nea(blat1, alpha1)

        sigma2, w_2, alpha2 = _solve_triangle_neb(cos_blat1, cos_blat2, sin_blat2,
                                                  sin_alpha0, alpha1)

        # Determine lamda12
        epsi = _normalize_equatorial_azimuth(cos_alpha0, e2m)
        fun_i3 = _get_i3_fun(epsi, eta)
        lamda1 = w_1 - f * sin_alpha0 * fun_i3(sigma1)  # Eq. 8
        lamda2 = w_2 - f * sin_alpha0 * fun_i3(sigma2)  # Eq. 8
        lamda12 = lamda2-lamda1

        # Update alpha1
        fun_j = _get_jfun(epsi)
        k_2 = e2m * cos_alpha0 ** 2
        sin_sigma1, cos_sigma1 = sin(sigma1), cos(sigma1)
        sin_sigma2, cos_sigma2 = sin(sigma2), cos(sigma2)
        k_sin_s1 = sqrt(1+k_2*sin_sigma1**2)
        k_sin_s2 = sqrt(1+k_2*sin_sigma2**2)
        delta_j = fun_j(sigma2) - fun_j(sigma1)
        m12 = b*(k_sin_s2*cos_sigma1*sin_sigma2
                 - k_sin_s1*cos_sigma2*sin_sigma1
                 - cos_sigma1*cos_sigma2*delta_j)  # Eq 38
        # M12 is (cos_sigma1 * cos_sigma2
        #        + k_sin_s2 / k_sin_s1 * sin_sigma1 * sin_sigma2
        #        - sin_sigma1 * cos_sigma2 * delta_j / k_sin_s1)  # Eq 39
        cos_alpha2 = cos(alpha2)
        dlamda12_dalpha1 = np.where(np.abs(cos_alpha2) < tol,
                                    -sqrt(1 - e_2 * cos_blat1**2) / sin_blat1 * 2,
                                    m12 / a / (cos_alpha2 * cos_blat2))
        dlamda12 = true_lamda12-lamda12

        dalpha1 = dlamda12/dlamda12_dalpha1

        return dalpha1, dlamda12

    dalpha_old = np.zeros_like(alpha1)
    for _ in range(20):
        dalpha1, dlamda12 = _newton_step(alpha1)
        if np.any(np.isnan(dalpha1)):
            dalpha_old *= 0.5
            alpha1 -= dalpha_old
        else:
            alpha1 = (alpha1 + dalpha1).clip(0, np.pi)  # alpha1 must be in [0, pi]
            dalpha_old = dalpha1
        if np.all(np.abs(dlamda12) < 1e-12):
            break
    else:
        warnings.warn("Max iterations reached. Newton method did not converge.",
                      stacklevel=2)
    return alpha1


def _canonical(blat1, blat2, lamda12):
    """Return canonical problem where blat1 <=0, blat1 <= blat2 <= -blat1 and 0<=lamda12<=pi."""
    blat1, blat2 = truncate_small(blat1), truncate_small(blat2)
    swap_bmask = np.abs(blat1) < np.abs(blat2)
    blat11, blat22 = np.where(swap_bmask, blat2, blat1), np.where(swap_bmask, blat1, blat2)
    negate_blat11 = blat11 > 0
    blat11[negate_blat11] *= -1
    blat22[negate_blat11] *= -1
    true_lamda = truncate_small(normalize_angle(lamda12))
    negate_lamda = true_lamda < 0
    true_lamda[negate_lamda] *= -1

    swap_alpha = np.logical_xor(swap_bmask, negate_blat11)

    def restore(alpha1, alpha2):
        """Restore computed values"""
        az1, az2 = np.where(swap_bmask, alpha2, alpha1), np.where(swap_bmask, alpha1, alpha2)
        az1, az2 = np.where(swap_alpha, np.pi-az1, az1), np.where(swap_alpha, np.pi-az2, az2)
        az1[negate_lamda] *= -1
        az2[negate_lamda] *= -1

        return normalize_angle(az1), normalize_angle(az2)

    return blat11, blat22, true_lamda,  restore


def distance(lat_a, lon_a, lat_b, lon_b, a=6378137, f=1.0 / 298.257223563, degrees=False):
    """
    Returns shortest surface distance between positions A and B on an ellipsoid.

    Parameters
    ----------
    lat_a, lon_a: real scalars or vectors of length m.
        latitude(s) and longitude(s) of position A [deg or rad].
    lat_b, lon_b: real scalars or vectors of length n.
        latitude(s) and longitude(s) of position B [deg or rad].
    a: real scalar, default WGS-84 ellipsoid.
        Semi-major axis of the Earth ellipsoid given in [m].
    f: real scalar, default WGS-84 ellipsoid.
        Flattening [no unit] of the Earth ellipsoid. If f==0 then spherical
        Earth with radius a is used in stead of WGS-84.
    degrees: bool
        True if latitudes, longitudes and azimuths are given in degress

    Returns
    -------
    distance:  real scalars or vectors of length max(m,n).
        Surface distance [m] from A to B on the ellipsoid
    azimuth_a, azimuth_b: real scalars or vectors of length max(m,n).
        direction [rad or deg] of line at position a and b relative to
        North, respectively.

    Notes
    -----
    Solves the inverse geodesic problem of finding the length and azimuths of the
    shortest geodesic between points specified by lat_a, lon_a, lat_b, lon_b on an ellipsoid.
    Keep :math:`|f| <= 1/50` for full double precision accuracy.

    Examples
    --------
    **"Surface distance"**
    ----------------------

    Find the surface distance sAB (i.e. great circle distance) between two
    positions A and B. The heights of A and B are ignored, i.e. if they don't have
    zero height, we seek the distance between the points that are at the surface of
    the Earth, directly above/below A and B.  Use Earth radius 6371e3 m.
    Compare the results with exact calculations for the WGS-84 ellipsoid.

    In geodesy this is known as "The second geodetic problem" or
    "The inverse geodetic problem" for a sphere/ellipsoid.

    Solution for a sphere:
      >>> import numpy as np
      >>> from karney.geodesic import rad, sphere_distance_rad, distance

      >>> latlon_a = (88, 0)
      >>> latlon_b = (89, -170)
      >>> latlons = latlon_a + latlon_b
      >>> r_Earth = 6371e3  # m, mean Earth radius
      >>> s_AB = sphere_distance_rad(*rad(latlons))[0]*r_Earth
      >>> s_AB = distance(*latlons, a=r_Earth, f=0, degrees=True)[0] # or alternatively

      >>> "Ex5: Great circle = {:5.2f} km".format(s_AB / 1000)
      'Ex5: Great circle = 332.46 km'

    Exact solution for the WGS84 ellipsoid:
      >>> s_12, azi1, azi2 = distance(*latlons, degrees=True)
      >>> "Ellipsoidal distance = {:5.2f} km".format(s_12 / 1000)
      'Ellipsoidal distance = 333.95 km'


    See also
    --------
    sphere_distance_rad
    reckon
    """

    if degrees:
        lat_a, lon_a, lat_b, lon_b = rad(lat_a, lon_a, lat_b, lon_b)

    s_ab, azimuth_a, azimuth_b = _distance(lat_a, lon_a, lat_b, lon_b, a, f)
    if degrees:
        azimuth_a, azimuth_b = deg(azimuth_a, azimuth_b)
    return s_ab, azimuth_a, azimuth_b


def _distance(lat_a, lon_a, lat_b, lon_b, a=6378137, f=1.0 / 298.257223563):
    """
    Returns shortest surface distance between positions A and B on an ellipsoid.

    Parameters
    ----------
    lat_a, lon_a: real scalars or vectors of length m.
        latitude(s) and longitude(s) of position A in radians.
    lat_b, lon_b: real scalars or vectors of length n.
        latitude(s) and longitude(s) of position B in radians.
    a: real scalar, default WGS-84 ellipsoid.
        Semi-major axis of the Earth ellipsoid given in [m].
    f: real scalar, default WGS-84 ellipsoid.
        Flattening [no unit] of the Earth ellipsoid. If f==0 then spherical
        Earth with radius a is used in stead of WGS-84.

    Returns
    -------
    distance:  real scalars or vectors of length max(m,n).
        Surface distance [m] from A to B on the ellipsoid
    azimuth_a, azimuth_b: real scalars or vectors of length max(m,n).
        direction [rad] of line at position a and b relative to
        North, respectively.

    Notes
    -----
    Solves the inverse geodesic problem of finding the length and azimuths of the
    shortest geodesic between points specified by lat1, lon1, lat2, lon2 on an ellipsoid.

    See also
    --------
    sphere_distance_rad

    """
    scalar_result = np.ndim(a) == 0
    broadcast = np.broadcast_arrays
    lat1, lon1, lat2, lon2, a, f = broadcast(*np.atleast_1d(lat_a, lon_a, lat_b, lon_b, a, f))
    blat1, blat2, true_lamda12, restore = _canonical(arctan((1 - f) * tan(lat1)),  # Eq 6
                                                     arctan((1 - f) * tan(lat2)),  # Eq 6
                                                     lon2 - lon1)
    b = polar_radius(a, f)
    eta = third_flattening(f)
    e_2, e2m = eccentricity2(f)

    # assume lat1<=0 and lat1 < lat2 < -lat1 and  0 <= true_lamda12 <= pi

    cos_blat1 = cos(blat1) + TINY
    sin_blat2, cos_blat2 = sin(blat2), cos(blat2)+TINY

    def vincenty():
        """See table 3 in Karney"""
        wbar = sqrt(1 - e_2 * (0.5 * (cos_blat1 + cos_blat2))**2)  # Eq. 48
        w12 = true_lamda12 / wbar
        # Determine sigma2-sigma1 on the auxiliary sphere:
        sigma12, alpha1, alpha2 = sphere_distance_rad(blat1, 0, blat2, w12)
        s12 = a * wbar * sigma12
        return s12, alpha1, alpha2, sigma12

    def _solve_astroid():
        """See table 4 in Karney"""
        delta = np.where(f == 0, 1, np.abs(f * np.pi * cos_blat1**2))

        x = (true_lamda12 - np.pi) * cos_blat1 / delta
        y = (blat1 + blat2) / delta

        alpha11 = _astroid(x, y, f)

        return alpha11

    def _solve_hybrid(alpha1):
        """See table 6 in Karney"""
        sigma1, w_1, cos_alpha0, sin_alpha0 = _solve_triangle_nea(blat1, alpha1)

        sigma2, w_2, alpha2 = _solve_triangle_neb(cos_blat1, cos_blat2, sin_blat2,
                                                  sin_alpha0, alpha1)

        # Determine s12:
        epsi = _normalize_equatorial_azimuth(cos_alpha0, e2m)

        i1fun = _get_i1_fun(epsi, return_inverse=False)

        s12 = b * np.abs((i1fun(sigma2) - i1fun(sigma1)))  # Eq. 7. I1(sigma2)

        fun_i3 = _get_i3_fun(epsi, eta)
        lamda1 = w_1 - f * sin_alpha0 * fun_i3(sigma1)  # Eq. 8
        lamda2 = w_2 - f * sin_alpha0 * fun_i3(sigma2)  # Eq. 8
        lamda12 = lamda2-lamda1
        if np.any(np.abs(lamda12 - true_lamda12) > 1e-8):
            warnings.warn("Some positions did not converge using newton method!....",
                          stacklevel=2)
        return s12, alpha2

    s12, alpha1, alpha2, sigma12 = vincenty()
    finite_mask = ~np.isnan(s12)
    if not np.any(finite_mask):
        return s12, alpha1, alpha2
    tol = 1e-12
    sin_lamda12 = sin(true_lamda12)
    sphere = (f == 0)

    meridional = (np.abs(sin_lamda12) <= tol)  # alpha1 is 0 or pi #lamda12
    delta_blat = blat2 - blat1
    equatorial = ((np.abs(delta_blat) <= tol)
                  & (np.abs(blat1) <= tol)
                  & (true_lamda12 <= (1-f)*np.pi))  # alpha1 is pi/2
    oblate = (f >= 0)
    prolate = (f < 0)
    mask = equatorial & ~(meridional & oblate)
    alpha1[mask] = np.pi/2
    alpha2[mask] = alpha1[mask]
    mask = meridional & ~(equatorial & prolate)
    alpha11 = np.sign(delta_blat)*true_lamda12
    alpha1[mask] = alpha11[mask]
    alpha2[mask] = true_lamda12[mask] - alpha1[mask]  # TODO check this
    nearly_antipodal = ~sphere & ~equatorial & (sigma12 >= np.pi*(1-3*np.abs(f)*cos_blat1**2))

    alpha1 = np.where(nearly_antipodal, _solve_astroid(), alpha1)

    short_distance = (s12 < a*1e-4)
    mask = ~(equatorial | meridional | short_distance | sphere) | nearly_antipodal
    mask *= finite_mask

    alpha1[mask] = _solve_alpha1(alpha1[mask], blat1[mask], blat2[mask], true_lamda12[mask], a, f)
    mask = ~(equatorial | short_distance | sphere) | nearly_antipodal
    mask *= finite_mask

    s12[mask], alpha2[mask] = _solve_hybrid(alpha1[mask])
    az1, az2 = restore(alpha1, alpha2)

    if scalar_result and s12.size == 1:
        return s12[0], az1[0], az2[0]
    return s12, az1, az2


_odict = globals()
__doc__ = (__doc__  # @ReservedAssignment
           + _make_summary(dict((n, _odict[n]) for n in __all__)))

if __name__ == "__main__":
    test_docstrings(__file__)




"""
Karney
------

This library provides native Python implementations of a subset of the C++ library, GeographicLib.

When given a combination of scalar and array inputs, the scalar inputs
are automatically expanded to match the shape of the arrays.

Full documentation is available at:
https://geographiclib.sourceforge.io/doc/library.html
https://github.com/geographiclib/geographiclib-octave


References
----------
C. F. F. Karney, "Algorithms for geodesics",
J. Geodesy 87, 43-55 (2013);
https://doi.org/10.1007/s00190-012-0578-z

"""

from . import geodesic, util, license
__version__ = "1.0.8"

__doc__ = (__doc__
           + geodesic.__doc__
           + util.__doc__
           + "License\n-------\n"
           + license.__doc__)






"""
Utility module
==============

"""
from __future__ import division, print_function
from collections import namedtuple
import numpy as np
from numpy import rad2deg, deg2rad
from karney._common import test_docstrings, _make_summary

__all__ = ["deg", "rad", "nthroot", "get_ellipsoid",
           "eccentricity2", "polar_radius", "third_flattening"]

FINFO = np.finfo(float)
_tiny_name = "tiny" if np.__version__ < "1.22" else "smallest_normal"
TINY = getattr(FINFO, _tiny_name)
EPS = FINFO.eps  # machine precision (machine epsilon)

Ellipsoid = namedtuple("Ellipsoid", "a f name")


ELLIPSOID = {
    1: Ellipsoid(a=6377563.3960, f=1.0 / 299.3249646, name="Airy 1858"),
    2: Ellipsoid(a=6377340.189, f=1.0 / 299.3249646, name="Airy Modified"),
    3: Ellipsoid(a=6378160.0, f=1.0 / 298.25, name="Australian National"),
    4: Ellipsoid(a=6377397.155, f=1.0 / 299.1528128, name="Bessel 1841"),
    5: Ellipsoid(a=6378249.145, f=1.0 / 293.465, name="Clarke 1880"),
    6: Ellipsoid(a=6377276.345, f=1.0 / 300.8017, name="Everest 1830"),
    7: Ellipsoid(a=6377304.063, f=1.0 / 300.8017, name="Everest Modified"),
    8: Ellipsoid(a=6378166.0, f=1.0 / 298.3, name="Fisher 1960"),
    9: Ellipsoid(a=6378150.0, f=1.0 / 298.3, name="Fisher 1968"),
    10: Ellipsoid(a=6378270.0, f=1.0 / 297, name="Hough 1956"),
    11: Ellipsoid(a=6378388.0, f=1.0 / 297,
                  name="Hayford/International ellipsoid 1924/European Datum 1950/ED50"),
    12: Ellipsoid(a=6378245.0, f=1.0 / 298.3, name="Krassovsky 1938"),
    13: Ellipsoid(a=6378145.0, f=1.0 / 298.25, name="NWL-9D / WGS 66"),
    14: Ellipsoid(a=6378160.0, f=1.0 / 298.25, name="South American 1969 / SAD69"),
    15: Ellipsoid(a=6378136.0, f=1.0 / 298.257, name="Soviet Geod. System 1985"),
    16: Ellipsoid(a=6378135.0, f=1.0 / 298.26, name="WGS 72"),
    17: Ellipsoid(a=6378206.4, f=1.0 / 294.9786982138, name="Clarke 1866 / NAD27"),
    18: Ellipsoid(a=6378137.0, f=1.0 / 298.257223563, name="GRS80 / WGS84 / NAD83"),
    19: Ellipsoid(a=6378137.0, f=298.257222101, name="ETRS89 / EUREF89"),
    20: Ellipsoid(a=6377492.0176, f=1/299.15281285, name="NGO1948")
}
ELLIPSOID_IX = {"airy1858": 1,
                "airymodified": 2,
                "australiannational": 3,
                "bessel": 4,
                "bessel1841": 4,
                "clarke1880": 5,
                "everest1830": 6,
                "everestmodified": 7,
                "fisher1960": 8,
                "fisher1968": 9,
                "hough1956": 10,
                "hough": 10,
                "hayford": 11,
                "international": 11,
                "internationalellipsoid1924": 11,
                "europeandatum1950": 11,
                "ed50": 11,
                "krassovsky": 12,
                "krassovsky1938": 12,
                "nwl-9d": 13,
                "wgs66": 13,
                "southamerican1969": 14,
                "sad69": 14,
                "sovietgeod.system1985": 15,
                "wgs72": 16,
                "clarke1866": 17,
                "nad27": 17,
                "grs80": 18,
                "wgs84": 18,
                "nad83": 18,
                "euref89": 19,
                "etrs89": 19,
                "ngo1948": 20
                }


def eccentricity2(f):
    """Returns the first and second eccentricity squared given the flattening, f.

    Parameters
    ----------
    f : array-like
        Flattening of the ellipsoid

    Notes
    -----
    The (first) eccentricity squared is defined as e2 = f*(2-f).
    The second eccentricity squared is defined as e2m = e2 / (1 - e2).
    """
    e2 = (2.0 - f) * f  # = 1-b**2/a**
    e2m = e2 / (1.0 - e2)
    return e2, e2m


def polar_radius(a, f):
    """Returns the polar radius b given the equatorial radius a and flattening f of the ellipsoid.

    Parameters
    ----------
    a : array-like
        Semi-major half axis or equatorial radius of ellipsoid
    f : array-like
        Flattening of the ellipsoid

    Notes
    -----
    The semi minor half axis (polar radius) is defined as b = (1 - f) * a
    where a is the semi major half axis (equatorial radius) and f is the flattening
    of the ellipsoid.

    """
    b = a * (1.0 - f)
    return b


def third_flattening(f):
    """Returns the third flattening, n, given the flattening, f.

    Parameters
    ----------
    f : array-like
        Flattening of the ellipsoid

    Notes
    -----
    The third flattening is defined as n = f / (2 - f).
    """

    return f / (2.0 - f)


def deg(*rad_angles):
    """
    Converts angle in radians to degrees.

    Parameters
    ----------
    rad_angles:
        angle in radians

    Returns
    -------
    deg_angles:
        angle in degrees

    Examples
    --------
    >>> import numpy as np
    >>> from karney.util import deg
    >>> deg(np.pi/2)
    90.0
    >>> deg(np.pi/2, [0, np.pi])
    (90.0, array([  0., 180.]))

    See also
    --------
    rad
    """
    if len(rad_angles) == 1:
        return rad2deg(rad_angles[0])
    return tuple(rad2deg(angle) for angle in rad_angles)


def rad(*deg_angles):
    """
    Converts angle in degrees to radians.

    Parameters
    ----------
    deg_angles:
        angle in degrees

    Returns
    -------
    rad_angles:
        angle in radians

    Examples
    --------
    >>> import numpy as np
    >>> from karney.util import deg, rad
    >>> deg(rad(90))
    90.0
    >>> deg(*rad(90, [0, 180]))
    (90.0, array([  0., 180.]))

    See also
    --------
    deg
    """
    if len(deg_angles) == 1:
        return deg2rad(deg_angles[0])
    return tuple(deg2rad(angle) for angle in deg_angles)


def nthroot(x, n):
    """
    Returns the n'th root of x to machine precision

    Parameters
    ----------
    x : real scalar or numpy array
    n : scalar integer

    Examples
    --------
    >>> import numpy as np
    >>> from karney.util import nthroot
    >>> np.allclose(nthroot(27.0, 3), 3.0)
    True

    """
    shape = np.shape(x)
    x = np.atleast_1d(x)
    y = x**(1. / n)
    mask = (x != 0) & (EPS * np.abs(x) < 1)
    ym = y[mask]
    y[mask] -= (ym**n - x[mask]) / (n * ym**(n - 1))
    if shape == ():
        return y[()]
    return y


def get_ellipsoid(name):
    """
    Returns semi-major axis (a), flattening (f) and name of reference ellipsoid as a named tuple.

    Parameters
    ----------
    name : string
        name of ellipsoid. Valid options are:
        1) Airy 1858
        2) Airy Modified
        3) Australian National
        4) Bessel 1841
        5) Clarke 1880
        6) Everest 1830
        7) Everest Modified
        8) Fisher 1960
        9) Fisher 1968
        10) Hough 1956
        11) Hayford / International ellipsoid 1924 / European Datum 1950 / ED50
        12) Krassovsky 1938
        13) NWL-9D / WGS 66
        14) South American 1969
        15) Soviet Geod. System 1985
        16) WGS 72
        17) Clarke 1866 / NAD27
        18) GRS80 / WGS84 / NAD83
        19) ETRS89 / EUREF89
        20) NGO1948

    Notes
    -----
    See also:
    https://en.wikipedia.org/wiki/Geodetic_datum
    https://en.wikipedia.org/wiki/Reference_ellipsoid


    Examples
    --------
    >>> from karney.util import get_ellipsoid
    >>> get_ellipsoid(name="wgs84")
    Ellipsoid(a=6378137.0, f=0.0033528106647474805, name="GRS80 / WGS84 / NAD83")
    >>> get_ellipsoid(name="GRS80")
    Ellipsoid(a=6378137.0, f=0.0033528106647474805, name="GRS80 / WGS84 / NAD83")
    >>> get_ellipsoid(name="NAD83")
    Ellipsoid(a=6378137.0, f=0.0033528106647474805, name="GRS80 / WGS84 / NAD83")
    >>> get_ellipsoid(name=18)
    Ellipsoid(a=6378137.0, f=0.0033528106647474805, name="GRS80 / WGS84 / NAD83")

    >>> wgs72 = get_ellipsoid(name="WGS 72")
    >>> wgs72.a == 6378135.0
    True
    >>> wgs72.f == 0.003352779454167505
    True
    >>> wgs72.name
    'WGS 72'
    >>> wgs72 == (6378135.0, 0.003352779454167505, "WGS 72")
    True
    """
    if isinstance(name, str):
        name = name.lower().replace(" ", "").partition("/")[0]
    ellipsoid_id = ELLIPSOID_IX.get(name, name)

    return ELLIPSOID[ellipsoid_id]


_odict = globals()
__doc__ = (__doc__  # @ReservedAssignment
           + _make_summary(dict((n, _odict[n]) for n in __all__)))


if __name__ == "__main__":
    test_docstrings(__file__)



import inspect
import textwrap
HAVE_INDENT = getattr(textwrap, "indent", None)
if HAVE_INDENT:
    def indent(text, amount=4, ch=" "):
        return textwrap.indent(text, amount * ch)
else:  # textwrap.indent function undefined (wasn't added until Python 3.3)
    def indent(text, amount=4, ch=" "):
        padding = amount * ch
        return "".join(padding+line for line in text.splitlines(True))

dedent = textwrap.dedent


def _get_h1line(object_):
    """Returns the H1 line of the documentation of an object."""
    doc = object_.__doc__
    if doc:
        return doc.partition("Parameters\n")[0].strip()
    return ""


def _make_summary(odict):
    """Return summary of all functions and classes in odict"""

    class_summary = "\n".join([":\n".join((oname, indent(_get_h1line(obj))))
                               for oname, obj in odict.items() if inspect.isclass(obj)])

    fun_summary = "\n".join([":\n".join((oname, indent(_get_h1line(obj))))
                             for oname, obj in odict.items() if not inspect.isclass(obj)])
    fmt = "{} in module\n{}----------\n{}\n\n"
    summary = ""
    if class_summary:
        summary = fmt.format("Classes", "-"*8, class_summary)
    if fun_summary:
        summary = summary + fmt.format("Functions", "-"*9, fun_summary)
    return summary


def use_docstring_from(cls):
    """This decorator modifies the decorated function's docstring by
    with the docstring from the class `cls`.

    If the function's docstring is None it is replaced with the supplied cls.__doc__.
    otherwise it is set to old_docstring.format(super=cls.__doc__)

    This is useful when you want to reuse the docstring from another class or
    if you want modify the docstring of a function at runtime.

    """
    return use_docstring(cls.__doc__)


def use_docstring(docstring):
    """This decorator modifies the decorated function's docstring with supplied docstring.

    If the function's docstring is None it is replaced with the supplied docstring.
    otherwise it is set to old_docstring.format(super=docstring)

    This is useful when you want modify the docstring of a function at runtime.
    """
    def _doc(func):
        func_docstring = func.__doc__
        if func_docstring is None:
            func.__doc__ = docstring
        else:
            try:
                new_docstring = dedent(func_docstring).format(super=docstring)
                func.__doc__ = new_docstring
            except Exception:
                pass  # python 2 crashes if the docstring alreasy exists!
        return func
    return _doc


def test_docstrings(filename):
    import doctest
    print("Testing docstrings in {0!s}".format(filename))
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS)
    print("Docstrings tested")


def write_readme(doc):
    with open("readme.txt", "w") as fid:
        fid.write(doc)
